@@ -36,8 +36,8 @@ int main(int argc, char * argv[]) {
    util_init();
    my_random_init(); // for opening book
 
-   printf("Toga II 1.4.1SE UCI based on Fruit 2.1 by Thomas Gaksch and Fabien Letouzey. Settings by Dieter Eberle\n");
-   printf("Experimental engine by Chris Formula. Code was based on Toga II 1.4 Beta5c by Thomas Gaksch\n");
+   printf("Toga II 3.0 UCI based on Fruit 2.1 by Thomas Gaksch, Fabien Letouzey and Jerry Donald. Settings by Dieter Eberle\n");
+   printf("Code was based on Toga II 1.4.1SE by Thomas Gaksch and Chris Formula, with bugfixes by Michel van den Bergh\n");
 
    // early initialisation (the rest is done after UCI options are parsed in protocol.cpp)
 
