@@ -38,19 +38,22 @@ static /* const */ int EndgameExchangePenalty = 30;
 
 static /* const */ int MaterialWeight = 256; // 100%
 
-static const int PawnOpening   = 70; // was 100 80
+static const int PawnOpening   = 75; /// J Donald's settings, was 80, 100, 70
 static const int PawnEndgame   = 90; // was 100
 static const int KnightOpening = 325;
-static const int KnightEndgame = 315; // Chris Formula's setting; was 325
+static const int KnightEndgame = 310; // J Donald's settings, was 315
 static const int BishopOpening = 325;
-static const int BishopEndgame = 315; // Chris Formula's setting; was 325
-static const int RookOpening   = 500;
-static const int RookEndgame   = 500;
+static const int BishopEndgame = 320; // Chris Formula's setting was 315
+static const int RookOpening   = 490; // J Donald's settings, was 500
+static const int RookEndgame   = 510; // J Donald's settings, was 500
 static const int QueenOpening  = 975; // was 1000
-static const int QueenEndgame  = 975; // was 1000
+static const int QueenEndgame  = 1000; 
 
-static const int BishopPairOpening = 50;
-static const int BishopPairEndgame = 70; // Chris Formula's setting; was 50
+static const int BishopPairOpening = 45; // JD setting; was 50
+static const int BishopPairEndgame = 65; // JD setting; was 50
+
+static /* const */ int KingPawnBonus = 30; // Endgame material adjustments
+static /* const */ int RookPawnPenalty = 10;
 
 // types
 
@@ -86,6 +89,10 @@ void material_parameter() {
    MaterialWeight = (option_get_int("Material") * 256 + 50) / 100;
    OpeningExchangePenalty = option_get_int("Toga Exchange Bonus");
    EndgameExchangePenalty = OpeningExchangePenalty; 
+   
+   // Endgame settings
+   KingPawnBonus = option_get_int("Toga King Pawn Endgame Bonus");
+   RookPawnPenalty = option_get_int("Toga Rook Pawn Endgame Penalty");
 
 }
 
@@ -126,6 +133,22 @@ void material_alloc() {
    }
 }
 
+// material_free()
+
+void material_free() {
+
+	int ThreadId;
+
+   ASSERT(sizeof(entry_t)==16);
+
+   if (UseTable) {
+
+		for (ThreadId = 0; ThreadId < NumberThreads; ThreadId++){
+		  my_free(Material[ThreadId]->table);
+		}
+   }
+}
+
 // material_clear()
 
 void material_clear(int ThreadId) {
@@ -722,9 +745,47 @@ static void material_comp_info(material_info_t * info, const board_t * board) {
       opening -= BishopPairOpening;
       endgame -= BishopPairEndgame;
    }
+   
+   // JD: King and Pawn Endgames (usually winning)
+   
+   if (wt - wp == 0 && bt - bp == 0){
+          
+      // TODO adjust by (wp-bp)*constant
+          
+      if (wp > bp+1){
+         endgame += KingPawnBonus*2;
+      }
+      else if (wp > bp){
+         endgame += KingPawnBonus;
+      }
+      
+      if (bp > wp+1){
+         endgame -= KingPawnBonus*2;
+      }
+      else if (bp > wp){
+         endgame -= KingPawnBonus;
+      }
+   }
+   
+   // Rook and Pawn Endgames (drawish)
+   if (wt - wp == 1 && bt - bp == 1 && wr == 1 && br == 1){
+      if (wp > bp){
+         endgame -= RookPawnPenalty; // note - sign
+      }
+      if (bp > wp){
+         endgame += RookPawnPenalty;
+      }
+   } 
+   
+   // Piece combo: Queen + Knight against Queen + Bishop
+   
+   if (wt-wp == 2 && bt-wp == 2 && wq == 1 && bq == 1 && wn == 1 && bb == 1)
+       endgame += 10;
+       
+   else if (wt-wp == 2 && bt-wp == 2 && wq == 1 && bq == 1 && wb == 1 && bn == 1)
+       endgame -= 10;
 
    // store info
-
    info->recog = recog;
    info->flags = flags;
    for (colour = 0; colour < ColourNb; colour++) info->cflags[colour] = cflags[colour];
