@@ -70,6 +70,9 @@ static HANDLE threat_handle[MaxThreads];
 static pthread_t threat_handle[MaxThreads];
 static my_sem_t thread_runnable[MaxThreads];
 #endif
+
+int NumberThreads = 1;
+
 //bool trans_endgame;
 //CRITICAL_SECTION CriticalSection; 
 
@@ -125,6 +128,7 @@ void search_clear() {
 		SearchInfo[ThreadId]->can_stop = false;
 		SearchInfo[ThreadId]->stop = false;
 		SearchInfo[ThreadId]->stopped = false;
+		SearchInfo[ThreadId]->exited = false;
 		SearchInfo[ThreadId]->check_nb = 10000; // was 100000
 		SearchInfo[ThreadId]->check_inc = 10000; // was 100000
 		SearchInfo[ThreadId]->last_time = 0.0;
@@ -210,6 +214,23 @@ void resume_threads(){
 }
 #endif
 
+// exit_threads() 
+
+void exit_threads() {
+  int ThreadId;
+  bool all_exited=false;
+  SearchInput->exit_engine=true; // hack
+  resume_threads();
+  while (!all_exited){
+    all_exited = true;
+    for (ThreadId = 1; ThreadId < NumberThreads; ThreadId++){
+      if (!SearchInfo[ThreadId]->exited)
+	all_exited = false;
+    }
+  }
+  SearchInput->exit_engine=false; // hack
+}
+
 
 // search()
 
@@ -301,13 +322,14 @@ void search() {
 unsigned __stdcall search_thread (void *param) {
 
 	int ThreadId = *((int*)param);
+	SearchInfo[ThreadId]->exited= false;
 
 	while (!SearchInput->exit_engine){
 		search_smp(ThreadId);
 		SearchInfo[ThreadId]->stopped = true;
 		SuspendThread(threat_handle[ThreadId-1]);
 	}
-
+	SearchInfo[ThreadId]->exited= true;
 	_endthreadex( 0 );
 	return 0;
 }
@@ -315,13 +337,14 @@ unsigned __stdcall search_thread (void *param) {
 void * search_thread (void *param) {
 
 	int ThreadId = *((int*)param);
+	SearchInfo[ThreadId]->exited= false;
 	my_sem_wait(&(thread_runnable[ThreadId-1]));
 	while (!SearchInput->exit_engine){
 	  search_smp(ThreadId);
 	  SearchInfo[ThreadId]->stopped = true;
 	  my_sem_wait(&(thread_runnable[ThreadId-1]));
 	}
-
+	SearchInfo[ThreadId]->exited= true;
 	return NULL;
 }
 #endif
@@ -332,6 +355,7 @@ void search_smp(int ThreadId) {
 
    int depth;
    int i;
+   int delta, a, b;
    bool search_ready;
    sint64 node_nb;
    double speed;
@@ -374,6 +398,7 @@ void search_smp(int ThreadId) {
    // iterative deepening
 
    search_ready = false;
+   delta = 16; 
 
    if (ThreadId == 0){ // main thread
 	   for (depth = 1; depth < DepthMax; depth++) {
@@ -386,20 +411,50 @@ void search_smp(int ThreadId) {
 			  SearchRoot[ThreadId]->change = false;
 
 			  board_copy(SearchCurrent[ThreadId]->board,SearchInput->board);
+			  
+			  // Aspiration windows (JD)
+			  
+			  if (depth <= 4){	// TODO: Try other values	  
+		         a = -ValueInf;
+                 b = +ValueInf;          
+              } 
+          
+              while (1){
 
-			  if (UseShortSearch && depth <= ShortSearchDepth) {
-				 search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,depth,SearchShort,ThreadId);
-			  } else {
-				 search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,depth,SearchNormal,ThreadId);
-			  }
+    		    if (UseShortSearch && depth <= ShortSearchDepth) {
+    		  	   search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,a,b,depth,SearchShort,ThreadId);
+    		    } else {
+    			   search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,a,b,depth,SearchNormal,ThreadId);
+    		    }
+    		    
+    		    // Aspiration windows
+    		    
+    		    // break on mate value
+    		    if (value_is_mate(SearchBest[ThreadId]->value)) break;
+    		    
+    		    // adjust on fail high/low
+    		    if (SearchBest[ThreadId]->value <= a){
+                    a -= delta;
+                    delta += delta;
+                } else if (SearchBest[ThreadId]->value >= b){
+                    b += delta;
+                    delta += delta;
+                } else {
+                    a = SearchBest[ThreadId]->value-delta;
+                    b = SearchBest[ThreadId]->value+delta; 
+                    break;
+                }    
+              }
 
 			  search_update_current(ThreadId);
 
 			  node_nb = 0;
 			  speed = 0;
+			  
+			  // calculate nodes/speed
 			  for (i = 0; i < NumberThreads; i++){
-				node_nb += SearchCurrent[ThreadId]->node_nb;
-				speed += SearchCurrent[ThreadId]->speed;
+				node_nb += SearchCurrent[i]->node_nb;
+				speed += SearchCurrent[i]->speed;
 			  }
 
 			  if (DispDepthEnd && SearchCurrent[ThreadId]->multipv == SearchInput->multipv) {
@@ -471,11 +526,34 @@ void search_smp(int ThreadId) {
 		  
 		  board_copy(SearchCurrent[ThreadId]->board,SearchInput->board);
 
-		  if (UseShortSearch && depth <= ShortSearchDepth) {
-			 search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,depth,SearchShort,ThreadId);
-		  } else {
-			 search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,depth,SearchNormal,ThreadId);
-		  }
+		  // Aspiration windows
+		  if (depth <= 4){	// Try other values	  
+	         a = -ValueInf;
+             b = +ValueInf;          
+          } 
+      
+          while (1){
+
+		    if (UseShortSearch && depth <= ShortSearchDepth) {
+		  	   search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,a,b,depth,SearchShort,ThreadId);
+		    } else {
+			   search_full_root(SearchRoot[ThreadId]->list,SearchCurrent[ThreadId]->board,a,b,depth,SearchNormal,ThreadId);
+		    }
+		    
+		    if (value_is_mate(SearchBest[ThreadId]->value)) break;
+		    
+		    if (SearchBest[ThreadId]->value <= a){
+                a -= delta;
+                delta += delta;
+            } else if (SearchBest[ThreadId]->value >= b){
+                b += delta;
+                delta += delta;
+            } else {
+                a = SearchBest[ThreadId]->value-delta;
+                b = SearchBest[ThreadId]->value+delta; 
+                break;
+            }    
+          }
 
 		  search_update_current(ThreadId);
 
@@ -501,10 +579,13 @@ void search_update_best(int ThreadId) {
 #else
    pthread_mutex_lock(&CriticalSection);
 #endif
-   if (DispBest && 
-	   (save_multipv[0].depth < SearchBest[ThreadId][0].depth ||
-		     (save_multipv[0].depth == SearchBest[ThreadId][0].depth &&
-			  save_multipv[0].value < SearchBest[ThreadId][0].value))) {
+
+/*   if (DispBest && 
+      (save_multipv[0].depth < SearchBest[ThreadId][0].depth || 
+           (save_multipv[0].depth == SearchBest[ThreadId][0].depth && 
+           save_multipv[0].value < SearchBest[ThreadId][0].value))) {*/ 
+
+     if (ThreadId == 0) {  // Norman Schmidt (kranium): multi-pv fix
 
       move = SearchBest[ThreadId][SearchCurrent[ThreadId]->multipv].move;
       value = SearchBest[ThreadId][SearchCurrent[ThreadId]->multipv].value;
