@@ -2,6 +2,8 @@
 
 // includes
 
+#include <math.h>
+
 #include "pawn.h"
 #include "attack.h"
 #include "board.h"
@@ -27,7 +29,7 @@
 
 #define ABS(x) ((x)<0?-(x):(x))
 #define MIN(X, Y)  ((X) < (Y) ? (X) : (Y))
-#define R_adpt(max_pieces, depth, Reduction) (Reduction + ((depth) > (9 + ((max_pieces<3) ? 2 : 0))))
+// #define R_adpt(max_pieces, depth, Reduction) (Reduction + ((depth) > (9 + ((max_pieces<3) ? 2 : 0))))
 
 
 // constants and variables
@@ -67,9 +69,8 @@ static const bool ExtendSingleReply = true; // true
 
 // razoring
 
-static const int RazorDepth = 3;
-static const int RazorMargin = 300;
-
+static const int RazorDepth = 1;
+static const int RazorMargin = 300; 
 
 // history pruning
 
@@ -78,9 +79,13 @@ static const int HistoryDepth = 3; // was 3
 static const int HistoryPVDepth = 3; // was 3
 static const int HistoryMoveNb = 3; // was 3
 static /* const */ int HistoryValue = 9830; // 60%
-static /* const */ int HistoryBound = 6400; // * 16384 + 50) / 100 10%=1638 15%=2458 20%=3277
-static /* const */ bool UseExtendedHistory = true;
-static const bool HistoryReSearch = true;
+//static /* const */ int HistoryBound = 6400; // * 16384 + 50) / 100 10%=1638 15%=2458 20%=3277
+//static /* const */ bool UseExtendedHistory = true;
+//static const bool HistoryReSearch = true;
+
+// Late Move Reductions (Stockfish/Protector)
+static int quietPvMoveReduction[64][64];       /* [depth][played move count] */
+static int quietMoveReduction[64][64]; /* [depth][played move count] */
 
 // futility pruning
 
@@ -95,8 +100,8 @@ static /* const */ int FutilityMargin3 = 950;
 static /* const */ bool UseDelta = true; // false
 static /* const */ int DeltaMargin = 50;
 
-static /* const */ int CheckNb = 1;
-static /* const */ int CheckDepth = 0; // 1 - CheckNb
+//static /* const */ int CheckNb = 1;
+//static /* const */ int CheckDepth = 0; // 1 - CheckNb
 
 // misc
 
@@ -193,8 +198,32 @@ void search_full_init(list_t * list, board_t * board, int ThreadId) {
    UseHistory = option_get_bool("History Pruning");
    HistoryValue = (option_get_int("History Threshold") * 16384 + 50) / 100;
 
-   UseExtendedHistory = option_get_bool("Toga Extended History Pruning");
-   HistoryBound = (option_get_int("Toga History Threshold") * 16384 + 50) / 100;
+   //UseExtendedHistory = option_get_bool("Toga Extended History Pruning");
+   //HistoryBound = (option_get_int("Toga History Threshold") * 16384 + 50) / 100;
+   
+   // Stockfish/Protector Late Move Reductions
+   // about 20 elo better than previous Toga History Reductions
+   int i, j;
+
+   for (i = 0; i < 64; i++){   
+      for (j = 0; j < 64; j++){
+         if (i == 0 || j == 0){
+            quietPvMoveReduction[i][j] = quietMoveReduction[i][j] = 0;
+         }
+         else{
+            double pvReduction =
+               log((double) (i)) * log((double) (j)) / 3.0;
+            double nonPvReduction =
+               0.33 + log((double) (i)) * log((double) (j)) / 2.25;
+            quietPvMoveReduction[i][j] =
+               (int) (pvReduction >= 1.0 ?
+                      floor(pvReduction) : 0);
+            quietMoveReduction[i][j] =
+               (int) (nonPvReduction >= 1.0 ?
+                      floor(nonPvReduction) : 0);
+         }
+      }
+   }
 
    // futility-pruning options
 
@@ -202,7 +231,6 @@ void search_full_init(list_t * list, board_t * board, int ThreadId) {
    FutilityMargin1 = option_get_int("Futility Margin");
    FutilityMargin2 = option_get_int("Extended Futility Margin");
 
-
    // delta-pruning options
 
    UseDelta = option_get_bool("Delta Pruning");
@@ -210,8 +238,8 @@ void search_full_init(list_t * list, board_t * board, int ThreadId) {
 
    // quiescence-search options
 
-   CheckNb = option_get_int("Quiescence Check Plies");
-   CheckDepth = 1 - CheckNb;
+   SearchCurrent[ThreadId]->CheckNb = option_get_int("Quiescence Check Plies");
+   SearchCurrent[ThreadId]->CheckDepth = 1 - SearchCurrent[ThreadId]->CheckNb;
 
    // standard sort
 
@@ -228,9 +256,9 @@ void search_full_init(list_t * list, board_t * board, int ThreadId) {
 
 // search_full_root()
 
-int search_full_root(list_t * list, board_t * board, int depth, int search_type, int ThreadId) {
+int search_full_root(list_t * list, board_t * board, int a, int b, int depth, int search_type, int ThreadId) {
 
-   int value, a, b;
+   int value;
 
    ASSERT(list_is_ok(list));
    ASSERT(board_is_ok(board));
@@ -243,17 +271,6 @@ int search_full_root(list_t * list, board_t * board, int depth, int search_type,
    ASSERT(board_is_legal(board));
    ASSERT(depth>=1);
 
-	a = SearchBest[ThreadId][SearchCurrent[ThreadId]->multipv].value - 30;
-	b = SearchBest[ThreadId][SearchCurrent[ThreadId]->multipv].value + 30;
-
-   if (SearchInput->multipv > 0){
-	   a = -ValueInf;
-      b = +ValueInf;
-   }
-
-   a = -ValueInf;
-   b = +ValueInf;
-
    value = full_root(list,board,a,b,depth,0,search_type, ThreadId);
 
    ASSERT(value_is_ok(value));
@@ -328,25 +345,10 @@ static int full_root(list_t * list, board_t * board, int alpha, int beta, int de
 	  //new_depth1 = full_new_depth(depth,move,board,board_is_check(board)&&LIST_SIZE(list)==1,false, height, ThreadId);
 
       move_do(board,move,undo);
-
+      
+      // search move. Changed by Jerry Donald to use aspiration windows (no Inf window researches needed)
       if (search_type == SearchShort || best_value[SearchCurrent[ThreadId]->multipv] == ValueNone) { // first move
 		  value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
-			if (value <= alpha){ // research
-				old_alpha = alpha = -ValueInf;
-				SearchRoot[ThreadId]->change = true;
-				SearchRoot[ThreadId]->easy = false;
-				SearchRoot[ThreadId]->flag = false;
-				search_update_root(ThreadId);
-			    value = -full_search(board,-beta,ValueInf,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
-			}
-		 else if (value >= beta){ // research
-			 SearchRoot[ThreadId]->change = true;
-			 SearchRoot[ThreadId]->easy = false;
-			 SearchRoot[ThreadId]->flag = false;
-			 search_update_root(ThreadId);
-			 value = -full_search(board,-ValueInf,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
-		 }
-  
       } else { // other moves
          value = -full_search(board,-alpha-1,-alpha,new_depth,height+1,new_pv,NodeCut,cap_extended,ThreadId);
          if (value > alpha) { // && value < beta
@@ -354,7 +356,7 @@ static int full_root(list_t * list, board_t * board, int alpha, int beta, int de
             SearchRoot[ThreadId]->easy = false;
             SearchRoot[ThreadId]->flag = false;
             search_update_root(ThreadId);
-			value = -full_search(board,-ValueInf,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
+			value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
          }
       }
 
@@ -432,6 +434,8 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
    int FutilityMargin;
    int probe_score, probe_depth;
    int newHistoryValue;
+   int threshold;
+   int reduction;
    entry_t * found_entry;
       
    ASSERT(board!=NULL);
@@ -447,9 +451,9 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 
    if (depth <= 0){
 	   if (node_type == NodePV)
-			CheckDepth = 1 - CheckNb - 1;
+			SearchCurrent[ThreadId]->CheckDepth = 1 - SearchCurrent[ThreadId]->CheckNb - 1;
 	   else
-			CheckDepth = 1 - CheckNb;
+			SearchCurrent[ThreadId]->CheckDepth = 1 - SearchCurrent[ThreadId]->CheckNb;
 	   return full_quiescence(board,alpha,beta,0,height,pv,ThreadId);
    }
 
@@ -466,7 +470,7 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
       search_check(ThreadId);
    }
 
-      // draw?
+    // draw?
 
 	if (board_is_repetition(board)) return ValueDraw;
 
@@ -486,8 +490,9 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 		         || board->cap_sq != SquareNone || PIECE_IS_PAWN(board->moving_piece))
 			  ) {
 		
-			if (probe_bitbases(board, probe_score)){
-				probe_score = value_from_trans(probe_score,height);
+            if (probe_bitbases(board, probe_score)){
+                if (probe_score < 0 && board_is_mate(board)) return VALUE_MATE(height);            
+                probe_score = value_from_trans(probe_score,height);
 
 			/*	trans_move = MoveNone;
 				trans_depth = depth;
@@ -500,6 +505,7 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 		}
 	}
 
+
 	if (recog_draw(board,ThreadId)) return ValueDraw;
 
    // mate-distance pruning
@@ -599,7 +605,9 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 
    // null-move pruning
 
-   if (UseNull && depth >= NullDepth && node_type != NodePV) {
+   if (UseNull && depth >= NullDepth && node_type != NodePV
+   && !(TRANS_IS_UPPER(trans_flags) && trans_depth >= depth - NullReduction - 1 && trans_value < beta)) { 
+   // if hash table tells us we are < beta, the null move search probably watses time
 
       if (!in_check
        && !value_is_mate(beta)
@@ -608,8 +616,8 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 
          // null-move search
 		 
-       //  if (depth < 11)
-			new_depth = depth - NullReduction - 1;
+        //  if (depth < 11)
+		 new_depth = depth - NullReduction - 1;
 		// else
 		//	new_depth = depth - NullReduction - 2;
 		 //new_depth = depth - R_adpt(board->piece_size[board->turn]+board->pawn_size[board->turn],depth,NullReduction) - 1;
@@ -617,7 +625,7 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 	     move_do_null(board,undo);
          value = -full_search(board,-beta,-beta+1,new_depth,height+1,new_pv,NODE_OPP(node_type),false,ThreadId);
          move_undo_null(board,undo);
-
+         
          // verification search
 
          if (UseVer && depth > VerReduction) {
@@ -655,11 +663,16 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
          }
 	  }
    }
-   // Razoring: idea by Tord Romstad (Glaurung)
-   else if(node_type != NodePV && depth <= RazorDepth &&  eval(board,alpha, beta, ThreadId) < beta - RazorMargin){
-		value = full_quiescence(board,alpha,beta,0,height,pv,ThreadId);
-      if(value < beta)
-        return value;
+    
+   // Razoring: idea by Tord Romstad (Glaurung), fixed by Jerry Donald
+   
+   if (node_type != NodePV && !in_check && trans_move == MoveNone && depth <= 3){ 
+        threshold = beta - RazorMargin - (depth-1)*39; // Values from Protector (Raimund Heid)
+        if (eval(board,threshold-1, threshold, ThreadId) < threshold){
+		   value = full_quiescence(board,threshold-1,threshold,0,height,pv,ThreadId); 
+           if (value < threshold) // corrected - was < beta which is too risky at depth > 1
+               return value;
+        }
    }
 
 
@@ -694,14 +707,13 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 	  // extensions
 
       new_depth = full_new_depth(depth,move,board,single_reply,node_type==NodePV, height, extended, &cap_extended, ThreadId);
-	  
+      
       // history pruning
 
       value = sort->value; // history score
-	  if (!in_check && depth <= 6 /*SearchCurrent[ThreadId]->act_iteration / 2*/ && node_type != NodePV 
+	  if (!in_check && depth <= 6 && node_type != NodePV 
 		  && new_depth < depth && value < 2 * HistoryValue / (depth + depth % 2) /*2*/ 
-		  && played_nb >= 1+/*sort->capture_nb*/+depth && !move_is_dangerous(move,board)){ 
-			//history_reset(move,board);
+		  && played_nb >= 1+depth && !move_is_dangerous(move,board)){ 
 			continue;
 	  }
 
@@ -713,7 +725,7 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 
             ASSERT(!move_is_check(move,board));
 
-            // optimistic evaluation as modelled by Chris Formula
+            // optimistic evaluation (Chris Formula)
 
             if (opt_value == +ValueInf) {
 				if (depth>=2){
@@ -722,7 +734,7 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 				else{
 					FutilityMargin = FutilityMargin1;
 				}
-				opt_value = eval(board,alpha,beta,ThreadId) + FutilityMargin;
+				opt_value = eval(board,alpha-FutilityMargin,alpha,ThreadId) + FutilityMargin;
 				ASSERT(opt_value<+ValueInf);
             }
 
@@ -741,27 +753,27 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
             }
          }
       } 
+      
+      // Late Move Reductions
 	  
+	  // init	  
 	  reduced = false;
+	  reduction = 0;
 
+      // lookup reduction      
 	  if (UseHistory) {
-			if (!in_check && new_depth < depth && played_nb >= HistoryMoveNb 
+			if (!in_check && new_depth < depth && played_nb >= HistoryMoveNb
 				&& depth >= HistoryDepth && !move_is_dangerous(move,board)) {
+                         
 					if (good_cap && !move_is_tactical(move,board)){
 						good_cap = false;
 					}
+					
 					if (!good_cap){
-						if (node_type != NodePV){
-							new_depth--;
-							reduced = true;
-						}
-						else{
-							if (played_nb>= HistoryMoveNb+3){
-								new_depth--;
-								reduced = true;
-							}
-						}
-					}
+                        reduction = (node_type == NodePV ? quietPvMoveReduction[depth<64 ? depth: 63][played_nb<64? played_nb: 63]:
+                                     quietMoveReduction[depth<64 ? depth: 63][played_nb<64? played_nb: 63]);
+                        if (reduction > 0)  reduced = true;
+                    }
 			}
 	  }
 
@@ -769,27 +781,36 @@ static int full_search(board_t * board, int alpha, int beta, int depth, int heig
 
 	  move_do(board,move,undo);
 
-      if (node_type != NodePV || best_value == ValueNone) { // first move
-		 value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NODE_OPP(node_type),cap_extended,ThreadId);
-      } else { // other moves
-		 value = -full_search(board,-alpha-1,-alpha,new_depth,height+1,new_pv,NodeCut,cap_extended,ThreadId);
-         if (value > alpha) { // && value < beta
-			value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
+      if (node_type != NodePV || best_value == ValueNone) { // first move or non-pv
+		 value = -full_search(board,-beta,-alpha,new_depth-reduction,height+1,new_pv,NODE_OPP(node_type),cap_extended,ThreadId);
+         
+         // The move was reduced and fails high; so we research with full depth
+         if (reduced && value >= beta){
+		        value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NODE_OPP(node_type),cap_extended,ThreadId);
+
+         }
+         
+      } else { // other moves (all PV children)
+      
+		 value = -full_search(board,-alpha-1,-alpha,new_depth-reduction,height+1,new_pv,NodeCut,cap_extended,ThreadId);
+         
+         // In case of fail high:
+               
+         // If reduced then we try a research with node_type = NodePV
+         if (value > alpha && reduced){  // && value < beta
+			value = -full_search(board,-beta,-alpha,new_depth-reduction,height+1,new_pv,NodePV,cap_extended,ThreadId);
+		    
+            // Still fails high! We research to full depth
+            if (reduced && value >= beta){	
+		        value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
+            }
+         
+         // If not reduced we research as a PV node   
+         } else if (value > alpha){	
+		    value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV,cap_extended,ThreadId);
          }
       }
 
-      // history-pruning re-search
-
-	  if (HistoryReSearch && reduced && value >= beta) {
-
-         ASSERT(node_type!=NodePV);
-         
-		 new_depth++;
-		 ASSERT(new_depth==depth-1);
-		 
-         value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NODE_OPP(node_type),cap_extended,ThreadId);
-      } 
-
       move_undo(board,move,undo);
 
       played[played_nb++] = move;
@@ -831,7 +852,7 @@ cut:
 
       good_move(best_move,board,depth,height,ThreadId);
 
-      if (best_value >= beta && !move_is_tactical(best_move,board)) {
+      if (best_value >= beta && !move_is_tactical(best_move,board)) { // check is 204b and d, e
 
 		 ASSERT(played_nb>0&&played[played_nb-1]==best_move);
 
@@ -840,7 +861,7 @@ cut:
 			ASSERT(move!=best_move);
 			history_bad(move,board,ThreadId);
 		 }
-	 
+
 		 history_good(best_move,board,ThreadId);
 		
       }
@@ -960,7 +981,9 @@ static int full_quiescence(board_t * board, int alpha, int beta, int depth, int
    sort_t sort[1];
    undo_t undo[1];
    mv_t new_pv[HeightMax];
-   int probe_score, probe_depth;
+   int probe_score, probe_depth;  
+   int trans_move, trans_depth, trans_flags, trans_value;
+   entry_t * found_entry;
 
    ASSERT(board!=NULL);
    ASSERT(range_is_ok(alpha,beta));
@@ -1029,6 +1052,36 @@ static int full_quiescence(board_t * board, int alpha, int beta, int depth, int
          if (value <= alpha) return value;
       }
    }
+   
+   // transposition table: added by Jerry Donald, about +20 elo self-play
+   
+   trans_move = MoveNone;
+   
+   if (UseTrans) { 
+
+      if (trans_retrieve(Trans,&found_entry,board->key,&trans_move,&trans_depth,&trans_flags,&trans_value)) {
+
+		 trans_value = value_from_trans(trans_value,height);
+		 
+		 /////////////////////
+		 // Hash hit scheme //
+		 /////////////////////
+		 
+		 // In case of exact hit, return if this is not a PV node qsearch
+		 // In case of upper bound and value <= alpha, return for all node types
+		 // In case of lower bound and value >= beta, return for all node types
+		 
+		 // @tried returning always (~ -10 elo, 4000 games)
+		 // @tried returning only in non-PV nodes (~ -5 elo, 4000 games)
+		 // @tried doing trans move first (~ - 5 elo, 4000 games)
+
+	 	 if ((UseExact && beta == alpha+1 && trans_value != ValueNone && TRANS_IS_EXACT(trans_flags)) 
+		 	 || (TRANS_IS_LOWER(trans_flags) && trans_value >= beta)
+			 || (TRANS_IS_UPPER(trans_flags) && trans_value <= alpha)) {
+			 return trans_value;
+		 } 
+	  }
+   }
 
    // more init
 
@@ -1081,13 +1134,13 @@ static int full_quiescence(board_t * board, int alpha, int beta, int depth, int
    if(cd < 0 && board->piece_size[board->turn] <= 5)
 	   cd++; */
 
-   sort_init_qs(sort,board,attack, depth>=CheckDepth /* depth>=cd */);
+   sort_init_qs(sort,board,attack, depth>=SearchCurrent[ThreadId]->CheckDepth /* depth>=cd */);
 
    while ((move=sort_next_qs(sort)) != MoveNone) {
 
 	  // delta pruning
 
-      if (UseDelta && beta == old_alpha+1) {
+      if (UseDelta && beta == old_alpha+1) { // i.e. non-PV
 
          if (!in_check && !move_is_check(move,board) && !capture_is_dangerous(move,board)) {
 
@@ -1145,6 +1198,21 @@ static int full_quiescence(board_t * board, int alpha, int beta, int depth, int
    }
 
 cut:
+    
+   // store result in hash table (JD)
+    
+   if (UseTrans) {
+
+      trans_move = best_move;
+      trans_depth = 0;
+      trans_flags = TransUnknown;
+      if (best_value > old_alpha) trans_flags |= TransLower;
+      if (best_value < beta) trans_flags |= TransUpper;
+      trans_value = value_to_trans(best_value,height);
+
+      trans_store(Trans,board->key,trans_move,trans_depth,trans_flags,trans_value);
+
+   }
 
    ASSERT(value_is_ok(best_value));
 
@@ -1185,6 +1253,8 @@ static int full_new_depth(int depth, int move, board_t * board, bool single_repl
 		//|| (in_pv &&  board->square[MOVE_TO(move)] != PieceNone256 && !extended && see_move(move,board) >= -100)
 		//|| (in_pv && mate_threat == true)
 		|| move_is_check(move,board) && (in_pv || see_move(move,board) >= -100)) {
+        // @tried no check extension in endgame (~ -5 elo)
+        // @tried no bad SEE check extension in PV (~ -5 elo)
 		return new_depth+1;
    }
    if (in_pv && PIECE_IS_PAWN(MOVE_PIECE(move,board))){
